/*
 * IPVS:        Round-Robin Scheduling module
 *
 * Version:     $Id: ip_vs_rr.c,v 1.5 1999/10/06 10:04:46 wensong Exp $
 *
 * Authors:     Wensong Zhang <wensong@iinchina.net>
 *              Peter Kese <peter.kese@ijs.si>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Fixes/Changes:
 *     Wensong Zhang            :     changed the ip_vs_rr_schedule to return dest
 *     Julian Anastasov         :     fixed the NULL pointer access bug in debugging
 *     Wensong Zhang            :     changed some comestics things for debugging
 *     Wensong Zhang            :     changed for the d-linked destination list
 *     Wensong Zhang            :     added the ip_vs_rr_update_svc
 *     Wensong Zhang            :     added any dest with weight=0 is quiesced 
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <net/ip_masq.h>
#ifdef CONFIG_IP_MASQUERADE_MOD
#include <net/ip_masq_mod.h>
#endif
#include <linux/sysctl.h>
#include <linux/ip_fw.h>
#include <net/ip_vs.h>
#include <net/ip_cs.h>

struct server 
{
  char *name;
  __u32 addr;
};

static struct server server_list[5]={
  
  {"ace",0x80c6c0c6},
  {"frodo",0x80c6c0c1},
  {"oblib",0x80c6c0c3},
  {"gandalf",0x80c6c0c2},
  {"default",0x00000000},
  {"reject",0xffffffff} 
};

int server_num=5;	

struct ip_sticky
{
  __u32 dest_ip;
  __u32 source_ip;
};

struct ip_sticky ip_sticky_list[500];
int sticky_num;
__u32 source_ip;

extern char *xml_tag_list[];
extern int tag_number;
extern int rule_configure(__u32 saddr, __u32 daddr, __u16 sport, __u16 dport, __u8 protocol);
extern __u32 server_ip_addr [SERVER_NUMBER];
extern  char url[256];
extern  char xml_tag_value[50][256];

static char req_data[5000];
static char xml_data[5000];
static char  node[20][50];
static char  string1[20],string2[20],string3[20];
static char sub_string[1000];
static char node_value[1000];


int url_extract(char * req_data)
{
  char *tail;
  int size;
  tail=strstr(req_data,"HTTP");
  
  /*printk("######data in url_extract\n");
    printk("######REQ_DATA=%s\n",req_data);
    printk("######TAIL=%s\n",tail);*/
  if((req_data!=NULL)&&(tail!=NULL))
    {
      size=tail-req_data;
      /*printk("#########size=%d\n",size);*/
      
      memset(url,0,size+1);
      
      memcpy(url,req_data,size);
      url[size+1]='\0';
      
      /*		printk("########url=%s  len=%d\n",url,strlen(url));
			printk("\n");*/
      return 1;
    }
  /*printk("#######url=NULL\n");*/
  url[0]='\0';
  return 0;
  
}


int xml_extract(char *req_data)
{
  
  char *tail;
  int size;
  tail=strstr(req_data,"<?xml version=");
  
  /*printk("######data in url_extract\n");
    printk("######REQ_DATA=%s\n",req_data);
    printk("######TAIL=%s\n",tail);*/
  if((req_data!=NULL)&&(tail!=NULL))
    {
      size=tail-req_data;
      /*printk("#########size=%d\n",size);*/
      
      memset(xml_data,0,size+1);
      
      memcpy(xml_data,req_data,size);
      xml_data[size+1]='\0';
      
      /*		printk("########url=%s  len=%d\n",url,strlen(url));
			printk("\n");*/
      return 1;
    }
  /*printk("#######url=NULL\n");*/
  xml_data[0]='\0';
  return 0;
  
}
void xml_set_value(char *xml_data,int index)                                     
{
  int i=0;
  int len;
  int node_num;
  
  char *start, *end;
  
  
  
  
  string1[0]='\0';
  
  strcat(string1,"<");
  string2[0]='\0';
  strcat(string2,">");
  string3[0]='\0';
  strcat(string3,"</");
  
  start=xml_tag_list[index];
  
  do {
    end=strstr(start,".");
    if((end!=NULL)&&(start!=NULL)){
      len=end-start;

      memset(node[i],0,len+1);
      memcpy(node[i],start,len);
      node[i][len+1]='\0';
      
      /*printk("##########node[%d]=%s len=%d node_value=\n",i,node[i],len);*/
    }
    else
      break;
    
    start=end+1;
    i++;
    /*printk("###########i=%d\n",i);	*/	
  }while(start!=NULL);
  
  node_num=i;
  
  
  /*printk("#############node_num=%d\n",node_num);*/
  memset(sub_string,0,strlen(req_data)+1);
  memcpy(sub_string,req_data,strlen(req_data));
  sub_string[strlen(req_data)+1]='\0';
  /*printk("#####sub_string=%s\n",sub_string);*/
  for(i=0;i<node_num;i++)
    {
      strcat(string1,node[i]);
      /*printk("###########string1=%s\n",string1);*/
      strcat(string1,string2);
      /*printk("###########string1=%s\n",string1);*/
      strcat(string3,node[i]);
      strcat(string3,string2);
      /*printk("###########string=%s\n",string3);*/
      start=strstr(sub_string,string1);
      /*printk("###########start=%s\n",start);*/
      
      end=strstr(sub_string,string3);
      /*printk("###########end=%s\n",end);*/
      if((start!=NULL)&&(end!=NULL)){
	
	len=end-start-strlen(string1);
	memset(node_value,0,len+1);
	
	memcpy(node_value,start+strlen(string1),len);
	node_value[len+1]='\0';
	/*printk("###########sub_value=%s\n",node_value);*/
	string1[0]='\0';
	
	strcat(string1,"<");
	string2[0]='\0';
	strcat(string2,">");
	string3[0]='\0';
	strcat(string3,"</");
	memset(sub_string,0,strlen(node_value)+1);
	memcpy(sub_string,node_value,strlen(node_value));
	sub_string[strlen(node_value)+1]='\0';
      }
      else{
	sub_string[0]='\0';
	break;
      }
      
    }
  
  if(sub_string!=NULL){
    
    memset(xml_tag_value[index],0,strlen(sub_string)+1);
    memcpy(xml_tag_value[index],sub_string,strlen(sub_string));
    xml_tag_value[index][strlen(sub_string)+1]='\0';
    
  }
  
  else 
    xml_tag_value[index][0]='\0';;
  
  
}

__u32 check_sticky(__u32 addr)
{
  int i;
/*printk("#####saddr=%08x sticky.source=%08x sticky.dest=%08x\n",addr,ip_sticky_list[sticky_num-1].source_ip,ip_sticky_list[sticky_num-1].dest_ip);*/
  for(i=0;i<sticky_num;i++){
    if(addr==ip_sticky_list[i].source_ip){
     /*printk("####sticky found on %08x\n",ip_sticky_list[i].dest_ip);*/                                     
      return ip_sticky_list[i].dest_ip;
    }
  }
  return 0;
}

__u32 module_rule_matching(struct sk_buff *skb)
{
  
  struct iphdr *iph=skb->nh.iph;
  int data_len;
  unsigned char *payload;
  
  
  int i;
  __u32 ret_addr;
  __u32 saddr,daddr;
  __u16 sport,dport;
  __u8 protocol;
  union ip_masq_tphdr h;
  
  MOD_INC_USE_COUNT;
  
  h.raw = (char*) iph + iph->ihl * 4;
  
  saddr=ntohl(iph->saddr);
  daddr=ntohl(iph->daddr);
  sport=ntohs(h.th->source);
  dport=ntohs(h.th->dest);
  protocol=iph->protocol;
  source_ip=saddr;
  
  if((ret_addr=check_sticky(saddr))>0)
{
  /*printk("######IP_STICKY addr=%08x\n",ret_addr);*/
    goto return_directly;
}
  /*printk("##############THIS IS MATCHING MODULE FUNCTION \n");*/
  
  payload=skb->data+iph->ihl*4+skb->h.th->doff*4;
  data_len=skb->len-(iph->ihl*4+skb->h.th->doff*4);
/* printk("skb->len=%d\n",skb->len); */
  if(data_len<0)
    return 0;
  
  memset(req_data,0,data_len+1);
  
  memcpy(req_data,payload,data_len);
  req_data[data_len+1]='\0';
/*printk("#####req_data=%s\n",req_data);*/
  url_extract(req_data);
  
 /*strcat(req_data,"<?xml version=> 
 <!--======= purhcase request example =======--> <purchase> <customerName>CCL</customerName> <customerID>111222333</customerID> <item> <productID>309121544</productID> <productName>IBM Thinkpad T21</productName> <unitPrice>5000</unitPrice> <noOfUnits>10</noOfUnits> <subTotal>50000</subTotal> </item> <item> <productID>309121538</productID> <productName>Intel wireless LAN PC Card</productName> <unitPrice>200</unitPrice> <noOfUnits>10</noOfUnits> <subTotal>2000</subTotal> </item> <totalAmount>52000</totalAmount> </purchase>"); */
  
  
  if(strstr(req_data,"<?xml version")!=NULL){
    xml_extract(req_data);
    /*printk("##req_data=%s\n",req_data);*/
    for(i=0;i<tag_number;i++)
      
      {
	
	xml_set_value(xml_data,i);
     
	/*printk("#####xml_tad_value[%d]=%s\n",i,xml_tag_value[i]);*/
       }  
  }
  
  else{
    for(i=0;i<tag_number;i++)
      
      {
	
	xml_tag_value[i][0]='\0';
     }  
  }
  
  ret_addr=rule_configure(saddr,daddr,sport,dport,protocol);
  
 return_directly:MOD_DEC_USE_COUNT;
  /*printk("####returned address=%08x\n",ret_addr);*/
  return ret_addr;
  
  
  
	
  
  
}



__u32 route_to(char * server_name, int sticky_flag)
     
{
  
  int i;
  __u32 dest_addr;
  for(i=0;i<server_num;i++){
    if(strcmp(server_name,server_list[i].name)==0){
	/*printk("#####server=%s\n",server_list[i].name);*/
      if(sticky_flag==1){
	/*printk("###### SERVER_ROUTE_TO=  STICKY  \n");*/
	ip_sticky_list[sticky_num].dest_ip=server_list[i].addr;
	ip_sticky_list[sticky_num].source_ip=source_ip;
	sticky_num++;
	/*printk("##### in route_to sticky_num=%d\n",sticky_num);*/
      }
      dest_addr=server_list[i].addr;
      return dest_addr;
    }
  }
  return -1;
  
  
  
}



void register_rule_matching()
     
{
  sticky_num=0;
  

  
  
  rule_matching=module_rule_matching;
}


void unregister_rule_matching()
{
  
  
  
  rule_matching=ip_cs_norule;
}



int init_module(void)
{
  printk("#############INSERT MATCHING MODIFIED\n");
  register_rule_matching();
  
  return 0;
}

void cleanup_module(void)
{
  unregister_rule_matching();
  
  printk("################REMOVED MATCHING MODIFIED\n");
}

